/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#ifndef __cGalep5Device_h
  #define __cGalep5Device_h

#include <memory>

extern const char gLibGalep5Version[];

/// \brief Use this constant to calculate the needed number of bus-cycles to wait the given time in nano seconds.
#define MCU_NS2CYCLES(x) ((x/200))
/// \brief Use this constant to calculate the needed number of bus-cycles to wait the given time in micro-seconds.
#define MCU_US2CYCLES(x) (x*5)
/// \brief Use this constant to calculate the needed number of bus-cycles to wait the given time in milli-seconds.
#define MCU_MS2CYCLES(x) (x*5000)

class cG5AbstractPSCtrl;
class cG5DeviceCtrl;
class cG5SDCardCtrl;
class cG5Config;
class cG5PDIOCtrl;
class cG5AdapterModuleCtrl;
class cG5PMUCtrl;
class cG5AbstractButtonCtrl;

/**
  \brief This class encapsulates all galep5 features. It supports all devices of the galep5 series.
  It may throw exceptions of type std::string on it's construction and on index errors.
*/
class cGalep5Device
  {
  public:
    cGalep5Device();
    ~cGalep5Device() throw();

    /// \brief Returns a reference to the galep5 configuration object.
    cG5Config& config();

    /// \brief Returns a reference to the power supply object.
    cG5AbstractPSCtrl& psCtrl();

    /// \brief Returns a reference to the adapter module ctrl. object.
    cG5AdapterModuleCtrl& adapterModuleCtrl();

    /// \brief Returns a reference to the device control object.
    cG5DeviceCtrl& deviceCtrl();

    /// \brief Returns true if a sd-card slot is installed.
    bool isSDCardInstalled() const;
    /// \brief Returns a reference to the SD-Card control object.
    cG5SDCardCtrl& sdCardCtrl();

    /// \brief Returns true if the device features a Power Management Unit. (Only Galep5M)
    bool isPMUInstalled() const;
    /// \brief Returns the PMU.
    cG5PMUCtrl& pmuCtrl();

    /// \brief Returns true if the device has a button of given index.
    bool hasButton(unsigned int = 0) const;
    /// \brief Returns the button control object.
    cG5AbstractButtonCtrl& button(unsigned int = 0);

    /// \brief Returns the number of PDIO (slots) installed.
    unsigned int numInstalledPDIOs() const;
    /**
      \brief Returns a reference to a io / cpd ctrl object for the given index.

      Index 0 concerns all installed FPGAs.
      Index 1 = the first slot.
      Index numInstalledPDIOs()+1 = last slot.
      Will throw an exception on index errors!
    */
    cG5PDIOCtrl& pdioCtrl(unsigned int iIndex);

    // << cpu >> //
    /// \brief Waits the given amount of bus-cycles.
    void waitBusCycles(unsigned int);
    /// \brief lockInterrupts();
    void lockInterrupts();
    /// \brief unlockInterrupts();
    void unlockInterrupts();

  private:
    class cGalep5DeviceImpl;
    std::auto_ptr<cGalep5DeviceImpl> mpImpl;
    /// \brief We keep the FPGA-Bus mapping address here to speed up it's access. (waitBusCycles() needs this address.)
    unsigned int miFPGAMapping;
    /// \brief We keep the information about enabled interrupts here due performance reasons. (Not in pimpl.)
    unsigned int miLockMask;
  };

#endif // __cGalep5Device_h
