/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#ifndef __cG5AbstractStateCtrl_h
  #define __cG5AbstractStateCtrl_h

#include <memory>

/**
  \brief This class represents the device state.
*/
class cG5AbstractStateCtrl
  {
  public:
    enum DeviceState
      {
        dstIdle,         /// Device is idle (maybe claimed or unclaimed)
        dstReady,        /// Device is ready for a job (claimed)
        dstBusyByHost,   /// Device is busy with a program job
        dstBusyByUser,   /// Device is busy by user (firmware-update, web-interface, sd-card, ...)
        dstWaitForUser,  /// Device waits for user interaction (Galep5-D only: The Button.)
        dstShortBusy,    /// Device is short busy. (Green + Red)
        dstSystemError   /// Fatal device error. (Maybe firmware or hardware error.)
      };

  public:
    cG5AbstractStateCtrl();
    virtual ~cG5AbstractStateCtrl() throw();

    /// \brief Returns the current device state.
    virtual DeviceState state() const = 0;
    /// \brief Returns \c true if the device is ready to perform an operation.
    virtual bool readyForJob() const = 0;

    /// \brief Tries to set the device state.
    virtual void setDeviceState(DeviceState) = 0;

    /// \brief Sets the device progress.
    virtual void setProgress(unsigned int) = 0;
    /// \brief Returns the progress.
    virtual unsigned int progress() const = 0;

    /// \brief Setup an error state.
    virtual void setDeviceError(unsigned int iDiagCode) = 0;
    /// \brief Unsets the error state.
    virtual void unsetDeviceError() = 0;
    /// \brief Returns \c true on error states.
    virtual bool isDeviceError() const = 0;
    /// \brief Returns \c the device error code.
    virtual unsigned int deviceErrorCode() const = 0;

    /// \brief Insert a success show cycle.
    virtual void showSuccess() = 0;
    /// \brief Insert a failure show cycle.
    virtual void showFailure() = 0;
    /// \brief Insert an ident show cycle.
    virtual void showIdent() = 0;

    virtual void buttonGreen(bool) = 0;
    virtual void buttonRed(bool) = 0;

    /**
       \brief Claim device.

       \param InterfaceName usb0 eth0 or what ever.
       \param sRemoteId location.toString() of remote machine.
       \param boForce If true a already claimed device will reassigned to new machine.

       If this operation will be called on a claimed galep and boForce is false an exception will be thrown.
    */
    virtual void claimDevice(const std::string& sInterfaceName, const std::string& sRemoteId, bool boForce) = 0;
    /// \brief Unclaims the device from remote machine and remove user.
    virtual void unclaimDevice() = 0;
    /// \brief Returns the interface id.
    virtual std::string claimedInterface() const = 0;
    /// \brief Returns the remote id.
    virtual std::string claimedRemoteId() const = 0;

    /// \brief Only for testing!
    virtual void setLED(char) = 0;
  };

#endif // __cG5AbstractStateCtrl_h
