/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <string>
#include "cG5Fw4xStateCtrl.h"
#include <cDriverAttribute.h>
#include <cG5Config.h>
#include <cG5ProductInfo.h>

#include <iostream>
#include <fstream>

class cGlobalDeviceState
  {
  public:
    /**
      \brief Creates an object representing the current device state.

      You may call update to make the state valid.
    */
    cGlobalDeviceState()
      {
        // Device starts unclaimed and idle.
        msClaimedRID = "";
        msClaimedIF = "";
        miCurrentState = cG5AbstractStateCtrl::dstIdle;
        load();
      }

    /**
      \brief Creates a new global state object but not stores them to operating system.
    */
    cGlobalDeviceState(const std::string& sClaimedIF, const std::string& sClaimedRID, cG5Fw4xStateCtrl::DeviceState iSt)
      : msClaimedIF(sClaimedIF), msClaimedRID(sClaimedRID), miCurrentState(iSt)
        { }

    /**
    */
    cGlobalDeviceState& operator=(const cGlobalDeviceState& with)
      {
        cGlobalDeviceState tmp(with);
        swap(tmp);
        return *this;
      }

    const std::string& claimedRemoteId() const
      { return msClaimedRID; }

    const std::string& claimedInterface() const
      { return msClaimedIF; }

    bool isClaimed() const
      { return (!(msClaimedRID.empty())); }


    /**
      \brief Reads the global device state from OS.
    */
    void load()
      {
        std::string sCIF;
        std::string sCRID;
        cG5Fw4xStateCtrl::DeviceState iSt = cG5Fw4xStateCtrl::dstIdle;

        std::ifstream s("/var/state", std::ios::in);
        if (s.is_open())
          { // if the file does not exists we assume a default //
            s >> sCIF;
            s >> sCRID;
            int i = 0;
            s >> i;
            iSt = static_cast<cG5Fw4xStateCtrl::DeviceState>(i);
            if (sCIF == "none")
              sCIF.erase();
            if (sCRID == "none")
              sCRID.erase();
          }


        msClaimedRID.swap(sCRID);
        msClaimedIF.swap(sCIF);
        miCurrentState = iSt;
      }

    /**
      \brief Writes the operating systems device state.

      (May fail if another process tries to write the state at the same time.)
      If this function failes the internal state is unchanged.
    */
    void store() const
      {
        std::ofstream s("/var/state", std::ios::out);
        if (!(s.is_open()))
          throw std::string("Unable to update device state!");

        std::string sCIF(msClaimedIF);
        std::string sCRID(msClaimedRID);
        cG5Fw4xStateCtrl::DeviceState iSt = miCurrentState;

        if (sCIF.empty())
          sCIF = "none";
        if (sCRID.empty())
          sCRID = "none";

        int i = static_cast<int>(iSt);
        s << sCIF;
        s << "\n";
        s << sCRID;
        s << "\n";
        s << i;
        s << "\n";
      }


    /// \brief Nonthrowing swap.
    void swap(cGlobalDeviceState& with)
      {
        msClaimedRID.swap(with.msClaimedRID);
        msClaimedIF.swap(with.msClaimedIF);
        std::swap(miCurrentState, with.miCurrentState);
      }

    cG5Fw4xStateCtrl::DeviceState state() const
      { return miCurrentState; }

  private:
    std::string msClaimedIF;
    std::string msClaimedRID;

    cG5Fw4xStateCtrl::DeviceState miCurrentState;
  };


class cG5Fw4xStateCtrl::cG5Fw4xStateCtrlImpl
  {
  public:
    cG5Fw4xStateCtrlImpl(const cG5Config& config)
      : mGlobalState() // reads current state of operating system //
      {
        mboInverted = false;
        msPowerLED = "1";
        msReadyLED = "2";
        msBusyLED = "3";
        
        switch (config.productInfo().productType())
          {
            case cG5ProductInfo::Galep5M:
              msPowerLED = "1";
              msReadyLED = "2";
              msBusyLED = "3";
              mboInverted = true;
             break;
            case cG5ProductInfo::Galep5Pro:
              msBusyLED = "2";
              msReadyLED = "1";
              msPowerLED = "3";
              mboInverted = true;
             break;
            case cG5ProductInfo::Galep5P:
              msPowerLED = "1";
              msReadyLED = "2";
              msBusyLED = "3";
              mboInverted = true;
             break;
            case cG5ProductInfo::Unknown:
             throw std::string ("cG5Fw4xStateCtrl: Unknwon device");
          }
        showState(mGlobalState);    // posit the current state of operating system //
      }

    ~cG5Fw4xStateCtrlImpl() throw()
      { }

    void setDeviceState(cG5Fw4xStateCtrl::DeviceState st)
      {
        cGlobalDeviceState newState(mGlobalState.claimedInterface(), mGlobalState.claimedRemoteId(), st);
        showState(newState);
        newState.store();            // try to publish the new state //
        mGlobalState.swap(newState); // never throws //
      }

    void setClaimed(const std::string& sIF, const std::string& sRID)
      {
        cGlobalDeviceState newState(sIF, sRID, globalState().state());
        showState(newState);          // show new state (may throw) //
        newState.store();            // try to publish the new state //
        mGlobalState.swap(newState); // never throws //
      }


    void setBusy(bool bo)
      {
        cDriverAttribute attr(std::string("/sys/bus/platform/drivers/galep5-ui/led")+msBusyLED);
        attr.set(bo ^ mboInverted);
      }
    void setReady(bool bo)
      {
        cDriverAttribute attr(std::string("/sys/bus/platform/drivers/galep5-ui/led")+msReadyLED);
        attr.set(bo ^ mboInverted);
      }
    void setPower(bool bo)
      {
        cDriverAttribute attr(std::string("/sys/bus/platform/drivers/galep5-ui/led")+msPowerLED);
        attr.set(bo ^ mboInverted);
      }

    void setProgress(unsigned int i)
      {
      }

    void buttonRed(bool bo)
      {
        cDriverAttribute attr(std::string("/sys/bus/platform/drivers/galep5-ui/led5"));
        attr.set(bo ^ mboInverted);
      }
    
    void buttonGreen(bool bo)
      {
        cDriverAttribute attr(std::string("/sys/bus/platform/drivers/galep5-ui/led4"));
        attr.set(bo ^ mboInverted);
      }

    cGlobalDeviceState& globalState()
      { return mGlobalState; }


    /**
      \brief Show the current state.

      Synchronizes the internal state with state posits by leds.
      \todo Support blinking in kernel!
    */
    void showState(const cGlobalDeviceState& state)
       {
        bool boPower = true;
        bool boBusy = false;
        bool boReady = false;
        switch (state.state())
          {
            case cG5AbstractStateCtrl::dstIdle:
            case cG5AbstractStateCtrl::dstReady:
              if (state.isClaimed())
                boReady = true;
              else
                boBusy = true;
             break;

            case cG5AbstractStateCtrl::dstShortBusy:
               boReady = true;
               boBusy = true;
             break;


            case cG5AbstractStateCtrl::dstSystemError:
              boPower = false;
              boBusy = true;
             break;

            case cG5AbstractStateCtrl::dstWaitForUser:
             break;

            case cG5AbstractStateCtrl::dstBusyByHost:
            case cG5AbstractStateCtrl::dstBusyByUser:
              boBusy = true;
             break;
          }

        setBusy(boBusy);
        setPower(boPower);
        setReady(boReady);
      }

  private:
    cGlobalDeviceState mGlobalState;

    bool mboInverted;
    std::string msPowerLED;
    std::string msReadyLED;
    std::string msBusyLED;
  };



cG5Fw4xStateCtrl::cG5Fw4xStateCtrl(const cG5Config& config)
  : cG5AbstractStateCtrl(), mpImpl(new cG5Fw4xStateCtrlImpl(config))
    { }

cG5Fw4xStateCtrl::~cG5Fw4xStateCtrl() throw()
  { /* auto_ptr destroys the implementation */ }

cG5Fw4xStateCtrl::DeviceState cG5Fw4xStateCtrl::state() const
  {
    return mpImpl->globalState().state();
  }

bool cG5Fw4xStateCtrl::readyForJob() const
  {
    return false;
  }

void cG5Fw4xStateCtrl::setDeviceState(DeviceState s)
  { mpImpl->setDeviceState(s); }


void cG5Fw4xStateCtrl::setProgress(unsigned int i)
  { mpImpl->setProgress(i); }

unsigned int cG5Fw4xStateCtrl::progress() const
  { return 0; }

void cG5Fw4xStateCtrl::setDeviceError(unsigned int)
  { }

void cG5Fw4xStateCtrl::unsetDeviceError()
  {
  }

bool cG5Fw4xStateCtrl::isDeviceError() const
  { return true; }

unsigned int cG5Fw4xStateCtrl::deviceErrorCode() const
  { return 0; }

void cG5Fw4xStateCtrl::showSuccess()
  { }

void cG5Fw4xStateCtrl::showFailure()
  { }

void cG5Fw4xStateCtrl::showIdent()
  { }

void cG5Fw4xStateCtrl::buttonGreen(bool bo)
  { mpImpl->buttonGreen(bo); }

void cG5Fw4xStateCtrl::buttonRed(bool bo)
  { mpImpl->buttonRed(bo); }


void cG5Fw4xStateCtrl::claimDevice(const std::string& sInterfaceName, const std::string& sRemoteId, bool boForce)
  {
    if ((mpImpl->globalState().isClaimed() && (!(boForce))))
      throw std::string("Device already claimed!");
    mpImpl->setClaimed(sInterfaceName, sRemoteId);
  }

void cG5Fw4xStateCtrl::unclaimDevice()
  {
    mpImpl->setClaimed("", "");
  }

std::string cG5Fw4xStateCtrl::claimedInterface() const
  { return mpImpl->globalState().claimedInterface(); }

std::string cG5Fw4xStateCtrl::claimedRemoteId() const
  { return mpImpl->globalState().claimedRemoteId(); }

void cG5Fw4xStateCtrl::setLED(char b)
  {
    mpImpl->setPower(b & 1);
    mpImpl->setReady(b & 2);
    mpImpl->setBusy(b & 4);
  }
