/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "cG5FPGACtrl.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
using namespace std;


/**
  Sollte boMappingOk false zurckgeben, darf keine Methode aufgerufen werden,
  da die FPGAs nicht im Arbeitsspeicher abgebildet wurden. (FPGA-Treiber nicht geladen)
*/
cG5FPGACtrl::cG5FPGACtrl(unsigned int iIdx, unsigned int iBaseMappingAddr)
  : mpIOSpace(0), miIndex(iIdx), miFPGABaseMapping(iBaseMappingAddr)
  {
    mpIOSpace = getFPGABase(iIdx);
  }

cG5FPGACtrl::~cG5FPGACtrl()
  {
  }

bool cG5FPGACtrl::isMappingValid() const
  {
    return (miFPGABaseMapping != 0);
  }

unsigned int cG5FPGACtrl::index() const
  { return (miIndex); }

/**
  Configure the fpga with index (iFPGAIndex) and image data dbImage.
  Input
    iFPGAIndex - index of fpga(0..n)
    dbImage - data to configure
  Return Value
    sReturn empty     : no error, success
    sReturn not empty : error, error message in human readable text format
*/
void cG5FPGACtrl::loadHdw(unsigned int iFPGAIndex, const char* pImage, unsigned int iSize, std::string& sReturn)
  {
    sReturn = "no such device";
    std::stringstream ss; ss << "/dev/fpga/" << iFPGAIndex;
    int iFd = open(ss.str().c_str(), O_RDWR);
    if (iFd < 0)
      {
        sReturn = "unable to open device";
        return;
      }
    if (write(iFd, pImage, iSize) != (signed)iSize)
      {
        int iErrNo = errno;
        std::cerr << "FPGA - device close return value: " << errno << std::endl;
        sReturn = "unable to configure device";
        switch (iErrNo)
          {
            case 5:  // internal device driver problem //
              sReturn += " (device driver internal problem)";
             break;
            case 10: // FPGA-reset init low problem //
              sReturn += " (init not low)";
             break;
            case 11: // FPGA-reset init high problem //
              sReturn += " (init not high)";
             break;
            case 20: // FPGA-config CRC error //
              sReturn += " (crc error while image upload)";
             break;
            case 21: // FPGA-config not ready //
              sReturn += " (fpga not ready)";
             break;
            case 30: // FPGA Id is wrong //
              sReturn += " (wrong hardware id)";
             break;
            case 31: // FPGA enumeration failed //
              sReturn += " (unable to enumerate / remap fpga)";
             break;
            default:
              sReturn += " (unknown problem [bug])";
          }
      }
    else
      sReturn = "";
    close(iFd);
  }



void cG5FPGACtrl::write16_volatile(unsigned int iOffset, unsigned short iData)
  {
    volatile unsigned short i = iData;
    mpIOSpace[iOffset] = i;
  }

/**
  Ungesicherte Lesefunktion.
  <br><b>Hinweis</b><br>
  Wenn im Konstruktor der FPGA-Index 0 bergeben wird, kann auf den gesamten FPGA-Speicher
  zugegriffen werden (unabhngig von internen, FPGA-Slot abhngigen mappings)
*/
unsigned short cG5FPGACtrl::read16_volatile(unsigned int iOffset)
  {
    volatile unsigned short i = mpIOSpace[iOffset];
    return (i);
  }


extern "C" void mcu_write32(unsigned int, unsigned int);
extern "C" unsigned int mcu_read32(unsigned int);
extern "C" void mcu_write16(unsigned int, unsigned int);
extern "C" unsigned int mcu_read16(unsigned int);

unsigned int cG5FPGACtrl::read32(unsigned int iOffs)
  { return mcu_read32((iOffs<<2)+(unsigned int)mpIOSpace); }

void cG5FPGACtrl::write32(unsigned int iOffs, unsigned int i)
  { mcu_write32((iOffs<<2)+(unsigned int)mpIOSpace, i); }


void cG5FPGACtrl::write16(unsigned int iOffset, unsigned short iData)
  { mcu_write16((iOffset <<1)+(unsigned int)mpIOSpace, iData); }

unsigned short cG5FPGACtrl::read16(unsigned int iOffset)
  { return (mcu_read16((iOffset<<1) + (unsigned int)mpIOSpace)); }



/**
  Legt Zuordnung zwischen einem Index (1..6) der FPGA-Platine und Adressraum des FPGAs fest.
  Der Wert 0 ist reserviert fr den Beginn des eingeblendeten Speichers.
*/
unsigned short* cG5FPGACtrl::getFPGABase(unsigned int i)
  {
    unsigned short* p = (unsigned short*)miFPGABaseMapping + (i*0x100);
    return (p);
  }
