/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <iostream>
#include <sstream>
#include "cG5PSCtrlR1.h"
#include "cSPICtrl.h"


#include <cGuardedBuffer.h>
#include <cTinyFsFileAttribute.h>

cG5PSCtrlR1::cCalibItem::cCalibItem(double dGradient, double dOffset)
  : mdGradient(dGradient), mdOffset(dOffset), mboValid(true)
    { }

cG5PSCtrlR1::cCalibItem::cCalibItem()
  : mdGradient(0), mdOffset(0), mboValid(false)
    { }

cG5PSCtrlR1::cCalibItem::cCalibItem(const cCalibItem& from)
  : mdGradient(from.mdGradient), mdOffset(from.mdOffset), mboValid(from.mboValid)
    { }

cG5PSCtrlR1::cCalibItem::cCalibItem& cG5PSCtrlR1::cCalibItem::operator=(const cCalibItem& from)
  {
    mdGradient = from.mdGradient;
    mdOffset = from.mdOffset;
    mboValid = from.mboValid;
    return *this;
  }

#include <math.h>
#include <iostream>
unsigned int cG5PSCtrlR1::cCalibItem::dacValue(unsigned int imV) const
  {
    unsigned long long dmV = imV;
//     std::cerr << "dGradient: " << mdGradient << std::endl;
//     std::cerr << "dOffset:   " << mdOffset << std::endl;
//     std::cerr << "dmV:       " << dmV << std::endl;
//     std::cerr << "dDAC:      " << (double(dmV)/mdGradient)+mdOffset << std::endl;
    double y = ((double(dmV)-mdOffset))/mdGradient;
//    std::cerr << "DAC : " << y << std::endl;
    int iResult = lround(y);
    if (iResult < 0)
      if (imV)
        iResult = 1;
      else
        iResult = 0;
    if (iResult > 1023)
      iResult = 1023;
    return (iResult);
  }

bool cG5PSCtrlR1::cCalibItem::valid() const
  { return (mboValid); }

void cG5PSCtrlR1::cCalibItem::invalidate()
  { mboValid = false; }



cG5PSCtrlR1::cG5PSCtrlR1()
  {
    mpSPICtrl = std::auto_ptr<cSPICtrl>(new cSPICtrl());
    loadCalibData();
  }

cG5PSCtrlR1::~cG5PSCtrlR1() throw()
  {
    // auto_ptr destroys the spi ctrl object //
  }




/**
  // TODO load order //
  \brief Loads calibration data.
    (1) - try load them from at-eeprom
    (2) - try laod them from /mnt/fs/cfg/ps-calib
    (3) - try load them from /mnt/fs/cfg/ps-calib.factory
    (4) - use uncalibrated values
*/
void cG5PSCtrlR1::loadCalibData()
  {
    std::auto_ptr<cGuardedBuffer> pBf;
    try {
      pBf = cTinyFsFileAttribute::getFileContent("ps-calib");
      if (pBf->size() < sizeof(cCalibItem)*wchNUM)
        {
          std::cout << "PSCtrl: Invalid calibration data (data too small)" << std::endl;
          return;
        }
    }
    catch (...)
      { // fall back to /mnt/fs/cfg/ps-calib //
        FILE* pCalibData = fopen("/mnt/fs/cfg/ps-calib", "rb");
        if (!(pCalibData))
          {
            std::cout << "PSCtrl: No calibration data found" << std::endl;
            return;
          }
        pBf.reset(new cGuardedBuffer(sizeof(cCalibItem)*wchNUM));
        if (fread(pBf->data(), 1, pBf->size(), pCalibData) != pBf->size())
          {
            std::cout << "PSCtrl: Calibration data invalid" << std::endl;
            fclose(pCalibData);
            return;
          }
        fclose(pCalibData);
      }

    cCalibItem* pCalibValues = reinterpret_cast<cCalibItem*>(pBf->data());

    // apply loaded calibration data //
    for (unsigned int i = 0; i < wchNUM; i++)
      {
        mCalibValues[i].invalidate();
        memcpy(&mCalibValues[i], &pCalibValues[i], sizeof(cCalibItem));
        if (!(mCalibValues[i].valid()))
          std::cerr << "PSCtrl - Warning: Supply " << i << " is not calibrated." << std::endl;
//        std::cout << "Supply: " << i << " Grad: " << mCalibValues[i].mdGradient << std::endl;
//        std::cout << "Offset: " << i << " Offs: " << mCalibValues[i].mdOffset << std::endl;
      }
  }

/**
  We store the calibration data as a tinyfs file.
*/
void cG5PSCtrlR1::storeCalibData()
  {
    cGuardedBuffer bf(sizeof(cCalibItem)*wchNUM);
    memcpy(bf.data(), &mCalibValues, bf.size());
    try {
      cTinyFsFileAttribute::setFileContent("ps-calib", bf);
    }
    catch(...)
      {
        FILE* pCalibData = fopen("/mnt/fs/cfg/ps-calib", "w+");
        if (!(pCalibData))
          {
            std::cout << "PSCtrl: Unable to creat calibration data file!" << std::endl;
            return;
          }
        if (fwrite(bf.data(), 1, bf.size(), pCalibData) != bf.size())
          {
            std::cout << "PSCtrl: Unable to write calibration data!" << std::endl;
            fclose(pCalibData);
            return;
          }
        fclose(pCalibData);
      }
  }



// << power supplies >> //

#define PS_READ_UIN_U_CHANNEL     0
#define PS_READ_VLS_I_CHANNEL     1
#define PS_READ_VCP2_U_CHANNEL    2
#define PS_READ_VCP2_I_CHANNEL    3
#define PS_READ_VCP1_U_CHANNEL    4
#define PS_READ_VCP1_I_CHANNEL    5
#define PS_READ_VCP3_U_CHANNEL    6
#define PS_READ_VCP3_I_CHANNEL    7

#define PS_WRITE_VCP3_U_CHANNEL   1
#define PS_WRITE_VGATE_U_CHANNEL  2
#define PS_WRITE_VCOMP_U_CHANNEL  3
#define PS_WRITE_RES1_U_CHANNEL   4
#define PS_WRITE_VLS_U_CHANNEL    5
#define PS_WRITE_RES2_U_CHANNEL   6
#define PS_WRITE_VCP2_U_CHANNEL   7
#define PS_WRITE_VCP1_U_CHANNEL   8


#define PS_READ_VLS_U_CHANNEL     5
#define PS_READ_VCOMP_U_CHANNEL   3
#define PS_READ_VGATE_U_CHANNEL   7

/**

*/
unsigned int uReadChannelToAtChannel(cG5PSCtrlR1::PSReadChannel i)
  {
    switch (i)
      {
        case cG5PSCtrlR1::rchVCP1:
          return (PS_READ_VCP1_U_CHANNEL);
        case cG5PSCtrlR1::rchVCP2:
          return (PS_READ_VCP2_U_CHANNEL);
        case cG5PSCtrlR1::rchVCP3:
          return (PS_READ_VCP3_U_CHANNEL);
        case cG5PSCtrlR1::rchVIn:
          return (PS_READ_UIN_U_CHANNEL);
        case cG5PSCtrlR1::rchVLS:
          return (PS_READ_VLS_U_CHANNEL);
        case cG5PSCtrlR1::rchVComp:
          return (PS_READ_VCOMP_U_CHANNEL);
        case cG5PSCtrlR1::rchVGate:
          return (PS_READ_VGATE_U_CHANNEL);

        case cG5PSCtrlR1::rchILS:
          return (PS_READ_VLS_I_CHANNEL);
        case cG5PSCtrlR1::rchICP1:
          return (PS_READ_VCP1_I_CHANNEL);
        case cG5PSCtrlR1::rchICP2:
          return (PS_READ_VCP2_I_CHANNEL);
        case cG5PSCtrlR1::rchICP3:
          return (PS_READ_VCP3_I_CHANNEL);
      }
    throw std::string("psCtrl: Invalid read channel!");
  }

unsigned int uWriteChannelToAtChannel(cG5PSCtrlR1::PSWriteChannel i)
  {
    switch (i)
      {
        case cG5PSCtrlR1::wchVCP1:
          return (PS_WRITE_VCP1_U_CHANNEL);
        case cG5PSCtrlR1::wchVCP2:
          return (PS_WRITE_VCP2_U_CHANNEL);
        case cG5PSCtrlR1::wchVCP3:
          return (PS_WRITE_VCP3_U_CHANNEL);
        case cG5PSCtrlR1::wchVLS:
          return (PS_WRITE_VLS_U_CHANNEL);
        case cG5PSCtrlR1::wchVGATE:
          return (PS_WRITE_VGATE_U_CHANNEL);
        case cG5PSCtrlR1::wchVCOMP:
          return (PS_WRITE_VCOMP_U_CHANNEL);
        case cG5PSCtrlR1::wchNUM: break;
      }
    throw std::string("psCtrl: Invalid write channel!");
  }

unsigned int readUADC(cG5PSCtrlR1::PSReadChannel i, cSPICtrl& mgr)
  {
    char bCh = uReadChannelToAtChannel(i);
    char bMSB = 0;
    char bLSB = 0;
    char bAck = 0;
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x71, false);
    mgr.spiIO(bCh, false);
    bMSB = mgr.spiIO(0, false);
    bLSB = mgr.spiIO(0, false);
    bAck = mgr.spiIO(0, false);
    mgr.setSelect(false);
    if (bAck != (char)0x95)
      {
        std::string sError("psCtrl: Invalid response while reading dac value response (0x");
        std::stringstream ss;
        ss << sError << std::hex << (unsigned int)bAck << ") - should be 0x95";
        throw std::string(ss.str());
      }
    unsigned int iReturn = bMSB; iReturn <<= 8; iReturn |= bLSB;
    return (iReturn);
  }

void writeUDAC(cG5PSCtrlR1::PSWriteChannel i, unsigned int iValue, cSPICtrl& mgr)
  {
    char bCh = uWriteChannelToAtChannel(i);
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x72, false);
    mgr.spiIO(bCh, false);
    mgr.spiIO((iValue >> 8) & 0xff, false);
    mgr.spiIO(iValue & 0xff, false);
    char bAck = mgr.spiIO(0, false);
    mgr.setSelect(false);
    if (bAck != (char)0x95)
      {
        std::string sError("psCtrl: Invalid response while writing dac value (0x");
        std::stringstream ss;
        ss << sError <<  std::hex << (unsigned int)bAck << ") - should be 0x95";
        throw std::string(ss.str());
      }
  }

unsigned char getPSFlags(cSPICtrl& mgr)
  {
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x73, false);
    char bCh = 0;
    unsigned char bState = mgr.spiIO(0, false);
    bCh = mgr.spiIO(0, false);
    if (bCh != 0x95)
      {
        std::string sError("psCtrl: Invalid response while reading flags (0x");
        std::stringstream ss;
        ss << sError << std::hex << (unsigned int)bCh << ") - should be 0x32";
        throw std::string(ss.str());
      }
    return bState;
  }


void setPSFlags(cSPICtrl& mgr, unsigned char bState)
  {
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x74, false);
    mgr.spiIO(bState, false);
    bState = mgr.spiIO(0, false);
    if (bState != 0x95)
      {
        std::string sError("psCtrl: Invalid response while writing flags (0x");
        std::stringstream ss;
        ss << sError << std::hex << (unsigned int)bState << ") - should be 0x42";
        throw std::string(ss.str());
      }
  }

/*
unsigned char psReadCtrl(cSPICtrl& mgr)
  {
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x73, false);
    char bStatus = mgr.spiIO(0, false);
    char bAck = mgr.spiIO(0, false);
    mgr.setSelect(false);
    if (bAck != (char)0x95)
      throw std::string("psCtrl: Invalid response!");
    return (bStatus);
  }

void psWriteCtrl(cSPICtrl& mgr, unsigned char bStatus)
  {
    mgr.setSelect(false);
    mgr.setSelect(true);
    mgr.spiIO(0x74, false);
    mgr.spiIO(bStatus, false);
    char bAck = mgr.spiIO(0, false);
    mgr.setSelect(false);
    if (bAck != (char)0x95)
      throw std::string("psCtrl: Invalid response!");
  }
*/


bool cG5PSCtrlR1::sltPSSetVoltageDAC(PSWriteChannel i, unsigned int iDAC)
  {
    try {
//    piDBG("Setting voltage (DAC)... " << iDAC);
    writeUDAC(i, iDAC, *mpSPICtrl.get());
    }
    catch (...)
      { return(false); }
    return (true);
  }

void cG5PSCtrlR1::getFwVersion(std::string& s) const
  {
    if (!(mpSPICtrl.get()))
      {
        s = "error";
	return;
      }
    s = "";
    mpSPICtrl->setSelect(false);
    mpSPICtrl->setSelect(true);
    mpSPICtrl->spiIO(0x90, false);
    for (unsigned int i = 0; i < 32; i++)
      {
        char b = mpSPICtrl->spiIO(0x90, false);
        if (!(b))
          break;
        s += b;
      }
    mpSPICtrl->setSelect(false);
  }


/**
  // URef = 3.3V

  Calculation of factor for VCP1..3:
    ev(solve(UOut(R1,R2,R3,R4,UFBFixed,UCtrl)= Ua, UCtrl), R1=47, R2=0.482, R3=1.5, R4=10)*1.0
            5765*Ua
    UCtrl = -------
             44647

          (UCtrl*1023)
    DAC = ------------
             URef

    DAC = 40*Ua // approx. //

  Calculation for VGATE:
    solve(UOutValue(1.2, 27, 3.3, UCtrl, 1.19)[1], UCtrl)[1];
    UCtrl = -(2200*UOut-82943)/18000

          ((-(2200*UOut-82943)/18000)*1023)
    DAC = -------------------------------
                   URef

  Calculation for VLS:
     solve(UOutValue(15, 31.6,  25.4, UCtrl, 1.19)[1], UCtrl)[1];
     UCtrl = -(874793450*UOut-4529160587)/1088325700

           ((-(874793450*UOut-4529160587)/1088325700)*1023)
     DAC = ------------------------------------------------
                           URef

  Calculation for VComp:
         Ua*R2
    Uc = -----
         R1+R2

           Ua*R2
           -----*1023
           R1+R2
    DAC = ------------
            URef
    DAC approx 40*Ua

*/
unsigned int cG5PSCtrlR1::calcDACValue(PSWriteChannel ch, unsigned int imV)
  {
    // use calibration values if possible //
    if (mCalibValues[ch].valid())
      {
//        std::cout << "Using calibration values... for ch " << ch << " / imV: " << imV << std::endl;
        return (mCalibValues[ch].dacValue(imV));
      }
    else
      std::cerr << "Warning: No valid calibration can be used." << std::endl;

    int iDAC = 0;
    switch(ch)
      {
      case wchVCP1:
      case wchVCP2:
      case wchVCP3:
        iDAC = ((40*imV) / 1000);
       break;

      case wchVLS:
        {
          double d = (((-((double)874793450.*(double)(imV/1000.)-(double)4529160587.)/(double)1088325700)*(double)1023) / (double)3.3);
          unsigned int i = (unsigned int)d;
          iDAC = (i);
        }
       break;

      case wchVGATE:
        {
          double d = 0.01722222222222*(82943.-2200.*((double)imV / 1000.));
          iDAC = (unsigned int)d;
        }
       break;

      case wchVCOMP:
        iDAC = ((imV*40)/1000);
       break;
      case wchNUM:
        break;
      }
    if (iDAC > 1023)
      iDAC = 1023;
    if (iDAC < 0)
      iDAC = 0;
    return (iDAC);
  }

// solve ((x*URef/1023) = (UM*R2)/(R1+R2), UM);
//
#include <iostream>
unsigned int calcVoltageFromADCValue(double dR1, double dR2, double dURef, unsigned int iADC)
  {
//    std::cout << "ADC: " << iADC << std::endl;
    double dUM = (dURef*iADC*dR2+dURef*iADC*dR1) / (1023.*dR2);
    return (unsigned int)(dUM);
  }


unsigned int cG5PSCtrlR1::calcVoltageFromADC(PSReadChannel ch, unsigned int iADC) const
  {
    switch (ch)
      {
        case rchVCP1:
        case rchVCP2:
        case rchVCP3:
          return (calcVoltageFromADCValue(470000, 82000, 3300, iADC));
          return (iADC / 4)*100;
         break;

        case rchVLS:
          return calcVoltageFromADCValue(1500, 1960, 3300, iADC);
        case rchVComp:
          return calcVoltageFromADCValue(1500, 1960, 3300, iADC);
        case rchVGate:
          return calcVoltageFromADCValue(100000, 4700, 3300, iADC);

        case rchVIn:
        case rchILS:
        case rchICP1:
        case rchICP2:
        case rchICP3:
          return (0);
      }
    return (0);
  }

bool cG5PSCtrlR1::sltPSSetVoltage(PSWriteChannel ch, unsigned int imV)
  {
    unsigned int iDACValue = 0;
    iDACValue = calcDACValue(ch, imV);
    return (sltPSSetVoltageDAC(ch, iDACValue));
  }


bool cG5PSCtrlR1::sltPSGetUValuesADC(unsigned int& iUIn, unsigned int& iVCP1, unsigned int& iVCP2, unsigned int& iVCP3, unsigned int& iVLS, unsigned int& iVGate, unsigned int& iVComp) const
  {
    try {
      iUIn = readUADC(cG5PSCtrlR1::rchVIn, *mpSPICtrl);
      iVCP1 = readUADC(cG5PSCtrlR1::rchVCP1, *mpSPICtrl);
      iVCP2 = readUADC(cG5PSCtrlR1::rchVCP2, *mpSPICtrl);
      iVCP3 = readUADC(cG5PSCtrlR1::rchVCP3, *mpSPICtrl);
      iVLS = readUADC(cG5PSCtrlR1::rchVLS, *mpSPICtrl);
      iVComp = readUADC(cG5PSCtrlR1::rchVComp, *mpSPICtrl);
      iVGate = readUADC(cG5PSCtrlR1::rchVGate, *mpSPICtrl);
    }
    catch (const std::string& s)
      { std::cerr << "Unable to get U-Values... " << s << std::endl; return (false); }
    return (true);
  }

bool cG5PSCtrlR1::sltPSGetIValuesADC(unsigned int& iIGate, unsigned int& iIComp, unsigned int& iILS, unsigned int& iICP1, unsigned int& iICP2, unsigned int& iICP3)
  {
    try {
      iIGate = 0; // will not be measured in power supply - rev.1
      iIComp = 0; // will not be measured in power supply - rev.1
      iILS = readUADC(cG5PSCtrlR1::rchILS, *mpSPICtrl);
      iICP1 = readUADC(cG5PSCtrlR1::rchVCP1, *mpSPICtrl);
      iICP2 = readUADC(cG5PSCtrlR1::rchVCP2, *mpSPICtrl);
      iICP3 = readUADC(cG5PSCtrlR1::rchVCP3, *mpSPICtrl);
    }
    catch (const std::string& s)
      { std::cerr << "Unable to get I-Values... " << s << std::endl; return (false); }
    return (true);
  }


/*
58 Messwerte bei einem Strom von 182.7mA:
   642 + 594 + 549 + 592 + 521 + 596 + 534 + 549 + 622 + 611 + 538 + 633 + 591 + 590 + 597 + 654 + 585 + 532 + 587 + 597 + 594 + 594 + 637 + 600 + 540 + 582 + 595 + 601 + 598 + 598 + 595 + 571 + 660 + 647 + 639 + 598 + 602 + 638 + 607 + 655 + 597 + 593 + 652 + 596 + 602 + 684 + 681 + 584 + 678 + 652 + 492 + 585 + 654 + 595 + 629 + 643 + 591 + 596
Average: 602.9138


182.7mA          x
--------   <=>   ----
602.9138         adc

  (adc * 182.7)
  ------------- = x
     602.9138

*/
#define calcI(adc) (unsigned int)(((adc * 182.7) / 602.9138)*1000)

bool cG5PSCtrlR1::sltPSGetIValues(unsigned int& iIGate, unsigned int& iIComp, unsigned int& iILS, unsigned int& iICP1, unsigned int& iICP2, unsigned int& iICP3)
  {
    if (!(sltPSGetIValuesADC(iIGate, iIComp, iILS, iICP1, iICP2, iICP3)))
      return false;
    iIGate = calcI(iIGate);
    iIComp = calcI(iIComp);
    iILS = calcI(iILS);
    iICP1 = calcI(iICP1);
    iICP2 = calcI(iICP2);
    iICP3 = calcI(iICP3);
    return true;
  }

#include <iostream>
bool cG5PSCtrlR1::sltPSGetUValues(unsigned int& iUIn, unsigned int& iVCP1, unsigned int& iVCP2, unsigned int& iVCP3, unsigned int& iVLS, unsigned int& iVGate, unsigned int& iVComp) const
  {
    if (!(sltPSGetUValuesADC(iUIn, iVCP1, iVCP2, iVCP3, iVLS, iVGate, iVComp)))
      return (false);
    iVCP1 = calcVoltageFromADC(rchVCP1, iVCP1);
    iVCP2 = calcVoltageFromADC(rchVCP2, iVCP2);
    iVCP3 = calcVoltageFromADC(rchVCP3, iVCP3);

    iVLS = calcVoltageFromADC(rchVLS, iVLS);
    iVComp = calcVoltageFromADC(rchVComp, iVComp);
    iVGate = calcVoltageFromADC(rchVGate, iVGate);

    if (iVGate < 20)
      { iVLS = 0; iVComp = 0; iVGate = 0; } // Readback not supported! //

    // uIn will not be calibrated for now //
    double d = double(iUIn)*12.8/672.;
    d*=1000;
    iUIn = (unsigned int)d;
    return (true);
  }

void cG5PSCtrlR1::sltSetSupplyCalibData(PSWriteChannel i, double dGradient, double dOffset)
  {
    mCalibValues[i] = cCalibItem(dGradient, dOffset);
    storeCalibData();
  }

enum FlagChannelName
  {
    RUN_VLS =   (1 << 6),
    VCP3_DOWN = (1 << 5),
    RUN_VCP3 =  (1 << 4),
    VCP2_DOWN = (1 << 3),
    RUN_VCP2 =  (1 << 2),
    VCP1_DOWN = (1 << 1),
    RUN_VCP1 =  (1 << 0)
  };

void cG5PSCtrlR1::sltSetDischarge(PSWriteChannel ch, bool bo)
  {
    unsigned char bFlags = getPSFlags(*mpSPICtrl);
    switch(ch)
      {
      case wchVCP1:
        if (bo)
          bFlags |= VCP1_DOWN;
        else
          bFlags &= ~VCP1_DOWN;
       break;
      case wchVCP2:
        if (bo)
          bFlags |= VCP2_DOWN;
        else
          bFlags &= ~VCP2_DOWN;
       break;
      case wchVCP3:
        if (bo)
          bFlags |= VCP3_DOWN;
        else
          bFlags &= ~VCP3_DOWN;
       break;
      default:
       break;
      }
    setPSFlags(*mpSPICtrl, bFlags);
  }
