/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "cG5ProductInfo.h"
#include <cFileAttribute.h>
#include <cTinyFsFileAttribute.h>

#include <string>
#include <sstream>
#include <unistd.h>
#include <sys/fcntl.h>

#include <cDFDevice.h>

struct cFPGAStatusT
  {
    unsigned int miFlags;
    unsigned int miImageID;
  };


class cG5ProductInfo::cG5ProductInfoImpl
  {
  public:
    cG5ProductInfoImpl()
      : miType(Unknown), msName("Unknown"), miRevision(0), msSerial(""), msProdSerial(""),
        miNumPDsInstalled(0), miFPGABaseMapping(0)
      {
        // Find out which device we have to support. //
        // This information is a file-attribute. //
        miType = cG5ProductInfo::Unknown;
        msName = cFileAttribute::getStringAttr("/etc/cfg/product");
        try {
          miRevision = cFileAttribute::getUIntAttr("/etc/cfg/product-revision");
        }
        catch(...)
         { miRevision = 0; }

        if (msName.find("D") != std::string::npos)
          miType = cG5ProductInfo::Galep5Pro;
        if (msName.find("P") != std::string::npos)
          miType = cG5ProductInfo::Galep5P;
        if (msName.find("M") != std::string::npos)
          miType = cG5ProductInfo::Galep5M;
        if (miType == Unknown)
          throw std::string("cG5ProductInfo: Unknwon device type");

        // Find out how many pindrivers has been installed. //
        // We ask the fpga device driver to get this information. //
        std::string sMap;
        miNumPDsInstalled = getNumPindriversInstalled(sMap);

        // Find out the fpga base mapping. //
        miFPGABaseMapping = cFileAttribute::getHexUIntAttr("/sys/bus/platform/drivers/galep5-fpga/mapping");

        // find out the production number, serial number //
        // the serial number and the production number are placed in the last 64 bytes, 32 bytes of physical flash memory //
        {
          char pBf[64];
          cDFDevice device("/dev/mtd0");
          device.rawRead(device.mediumSize()-64, pBf, 64);
          {
            char* p = (char*)memmem(pBf, 9, "DEVICE_ID", 9);
            if (p == pBf)
              {
                // we have the device - id (production number) //
                msProdSerial = "";
                for (unsigned int i = 9; i < 64; ++i)
                  {
                    if (pBf[i] == 0)
                      {
                        msProdSerial += pBf+9;
                        break;
                      }
                  }
              }
            else
              msProdSerial = "unset";
           }

         if (!(cTinyFsFileAttribute::tryGetStringAttr(msSerial, "serial")))
           msSerial = "unset";
        }
      }

    ~cG5ProductInfoImpl() throw()
      { }

  private:
    /// \brief Returns the number of pindriver slots detected by device driver.
    unsigned int getNumPindriversInstalled(std::string& sMap)
      {
        cFPGAStatusT aRecord[10];
        int iFd = open("/proc/galep5-fpga/bin-status", O_RDONLY);

        if (iFd < 0)
          throw std::string("cG5ProductInfo: FPGA device driver not loaded!"); // No FPGA driver present //
        unsigned int iRead = read(iFd, &aRecord, sizeof(aRecord));
        if (!(close(iFd) == 0))
          throw std::string("cG5ProductInfo: Unexpected result from FPGA device driver.");
        unsigned int iNumPD = 0;
        sMap = "";
        std::stringstream ss;
        unsigned int iNumValid = iRead / sizeof(cFPGAStatusT);
        for (unsigned int i = 0; i < iNumValid; i++)
          {
            if (aRecord[i].miFlags & 1)
              {
                iNumPD++;
                if (i)
                  ss << ", ";
                ss << i;
              }
          }
        sMap = ss.str();
        return (iNumPD);
      }

  public:
    cG5ProductInfo::ProductType miType;
    std::string msName;
    unsigned int miRevision;
    std::string msSerial;
    std::string msProdSerial;
    unsigned int miNumPDsInstalled;
    unsigned int miFPGABaseMapping;
  };








cG5ProductInfo::cG5ProductInfo()
  {
    mpImpl = std::auto_ptr<cG5ProductInfoImpl>(new cG5ProductInfoImpl());
  }


cG5ProductInfo::~cG5ProductInfo() throw()
  { /* auto_ptr destroys implementation */ }


std::string cG5ProductInfo::productName() const
  { return mpImpl->msName; }

unsigned int cG5ProductInfo::productRevision() const
  { return mpImpl->miRevision; }

cG5ProductInfo::ProductType cG5ProductInfo::productType() const
  { return mpImpl->miType; }

const std::string& cG5ProductInfo::serialNumber() const
  { return mpImpl->msSerial; }

const std::string& cG5ProductInfo::productionNumber() const
  { return mpImpl->msProdSerial; }

unsigned int cG5ProductInfo::getNumPDsInstalled() const
  { return mpImpl->miNumPDsInstalled; }

unsigned int cG5ProductInfo::fpgaBaseMappingAddr() const
  {
    return mpImpl->miFPGABaseMapping;
  }
