/**
   Copyright (c) 2005,2006,2007,2008 by Conitec Datasysteme GmbH
   Written 2008 by Peer Georgi <georgi@conitec.com>

   License: GPL

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "cDFDevice.h"
#include "cTinyFsFile.h"
#include <sstream>
#include <unistd.h>
#include <fcntl.h>

cDFDevice::cDFDevice(const std::string& sDevice)
  : miFd(0), mpSectorBuffer(1056)
    {
      miFd = ::open(sDevice.c_str(), O_RDWR);
      if (miFd < 0)
        throw std::string("unable to open device");
    }

cDFDevice::~cDFDevice() throw()
  {
//    ::lseek(miFd, 0, SEEK_END);
    ::close(miFd);
  }

#include <iostream>
#define piDBG(x) { std::cerr << x << std::endl; }


void cDFDevice::rawRead(unsigned int iMediaOffset, char* pDst, unsigned int iSize)
  {
//    piDBG("rawReadSize: " << iSize);
    unsigned int iStartSector = iMediaOffset / sectorSize();

    // alignmment (before) //
    {
      unsigned int iPre = iMediaOffset % sectorSize();
      if (iPre)
        {
          unsigned int iPreSize = sectorSize()-iPre;
          if (iPreSize > iSize)
            iPreSize = iSize;
//           piDBG("PreSize: " << iPreSize);
          readSector(iStartSector, mpSectorBuffer.data(), 1);
          memcpy(pDst, mpSectorBuffer.data()+iPre, iPreSize);
          iSize -= iPreSize;
          pDst += iPreSize;
          iStartSector++;
        }
    }

    // continuous block read //
    unsigned int iContinuousSectors = iSize / sectorSize();
    if (iContinuousSectors)
      {
//          piDBG("Continuous blocks: " << iContinuousSectors << " sectors");
        readSector(iStartSector, pDst, iContinuousSectors);
        unsigned int iContinuousSize = sectorSize()*iContinuousSectors;
        iSize -= iContinuousSize;
        pDst += iContinuousSize;
        iStartSector += iContinuousSectors;
      }

    // read pending data
    if (iSize)
      {
        if (iSize > sectorSize())
          throw std::string("internal error while reading");
        unsigned int iLeft = iSize % sectorSize();
        if (iLeft)
          {
//             piDBG("LeftSize: " << iSize);
            readSector(iStartSector, mpSectorBuffer.data(), 1);
            memcpy(pDst, mpSectorBuffer.data(), iSize);
          }
      }
  }


void cDFDevice::rawWrite(unsigned int iMediaOffset, const char* pSrc, unsigned int iSize)
  {
    unsigned int iStartSector = iMediaOffset / sectorSize();

    // alignmment (before) //
    {
      unsigned int iPre = iMediaOffset % sectorSize();
      if (iPre)
        {
          unsigned int iPreSize = sectorSize()-iPre;
          if (iPreSize > iSize)
            iPreSize = iSize;
//          piDBG("PreSize: " << iPreSize);
          readSector(iStartSector, mpSectorBuffer.data(), 1);
          memcpy(mpSectorBuffer.data()+iPre, pSrc, iPreSize);
          writeSector(iStartSector, mpSectorBuffer.data(), 1);
          iSize -= iPreSize;
          pSrc += iPreSize;
          iStartSector++;
        }
    }

    // continuous block read //
    unsigned int iContinuousSectors = iSize / sectorSize();
    if (iContinuousSectors)
      {
        // piDBG("Continuous blocks: " << iContinuousSectors << " sectors");
        writeSector(iStartSector, pSrc, iContinuousSectors);
        unsigned int iContinuousSize = sectorSize()*iContinuousSectors;
        iSize -= iContinuousSize;
        pSrc += iContinuousSize;
        iStartSector += iContinuousSectors;
      }

    // read pending data
    if (iSize)
      {
        if (iSize > sectorSize())
          throw std::string("internal error while writing");
        unsigned int iLeft = iSize % sectorSize();
        if (iLeft)
          {
            // piDBG("LeftSize: " << iSize);
            readSector(iStartSector, mpSectorBuffer.data(), 1);
            memcpy(mpSectorBuffer.data(), pSrc, iSize);
            writeSector(iStartSector, mpSectorBuffer.data(), 1);
          }
      }
  }

unsigned int cDFDevice::sectorSize() const
  { return mpSectorBuffer.size(); }

void cDFDevice::writeSector(unsigned int iStart, const char* p, unsigned int iNumSectors)
  {
    unsigned int iSectorSize = sectorSize();
    off_t i = ::lseek(miFd, iStart*iSectorSize, SEEK_SET);
    if ((unsigned)i != (off_t)iStart*iSectorSize)
      throw std::string("invalid seek");
    i = ::write(miFd, p, iNumSectors*iSectorSize);
    if (i < 0)
      throw std::string("unable to write");;
    if ((unsigned)i != (off_t)iNumSectors*iSectorSize)
      throw std::string("incomplete write");
  }

void cDFDevice::readSector(unsigned int iStart, char* p, unsigned int iNumSectors)
  {
    unsigned int iSectorSize = sectorSize();
    off_t i = ::lseek(miFd, iStart*iSectorSize, SEEK_SET);
    if ((unsigned)i != (off_t)iStart*iSectorSize)
      throw std::string("invalid seek");
    i = ::read(miFd, p, iNumSectors*iSectorSize);
    if (i < 0)
      throw std::string("unable to read");
    if ((unsigned)i != (off_t)iNumSectors*iSectorSize)
      throw std::string("incomplete read");
  }


unsigned int cDFDevice::mediumSize() const
  { return (1056*8192); }

unsigned int cDFDevice::numSectors() const
  { return (8192); }

unsigned int cDFDevice::preferredBlockSize() const
  { return 1056; }
